package com.jprofiler.api.agent;

import com.jprofiler.agent.AgentConstants;
import com.jprofiler.agent.ControllerImpl;

import java.io.File;

/**
 * This is JProfiler's profiling API.
 * It is contained in {@code bin/agent.jar} in
 * your JProfiler installation directory. Call these static methods from your source code
 * to influence the run time behaviour of JProfiler.
 * <p>The agent API is also available as a dependency, please replace "VERSION" in the snippets below
 * with the corresponding JProfiler version:
 * </p>
 * <b>Maven</b>
 * <pre>{@code
 * <dependency>
 *   <groupId>com.jprofiler</groupId>
 *   <artifactId>jprofiler-agent</artifactId>
 *   <version>VERSION</version>
 * </dependency>
 *
 * <repository>
 *   <id>ej-technologies</id>
 *   <url>http://maven.ej-technologies.com/repository</url>
 * </repository>
 * }</pre>
 * <b>Ivy</b>
 * <pre>{@code
 * <dependency org="com.jprofiler" name="jprofiler-agent" rev="VERSION" />
 *
 * <ibiblio name="ej-technologies" usepoms="true" m2compatible="true"
 * root="http://maven.ej-technologies.com/repository" />
 * }</pre>
 * <b>Gradle</b>
 * <pre>{@code
 * repositories {
 *   maven {
 *     url 'http://maven.ej-technologies.com/repository'
 *   }
 * }
 * dependencies {
 *   classpath group: 'com.jprofiler', name: 'jprofiler-agent', version: 'VERSION'
 * }
 * }</pre>
 * <p>
 * When executing your application without JProfiler, all calls into this class will
 * quietly do nothing.
 * </p>
 * <p>If you would like to interactively access the functionality in this class, please use the
 * {@code jpcontroller} command line executable that attaches to the JProfiler MBean which is registered
 * for every profiled application. For more information, please invoke {@code jprofiler -help}.
 *
 */
public class Controller {

    /**
     * Name of the built-in persistence probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_PERSISTENCE = AgentConstants.PROBE_NAME_PERSISTENCE;

    /**
     * Name of the built-in JDBC probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_JDBC = AgentConstants.PROBE_NAME_JDBC;

    /**
     * Name of the built-in JMS probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_JMS = AgentConstants.PROBE_NAME_JMS;

    /**
     * Name of the built-in JNDI probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_JNDI = AgentConstants.PROBE_NAME_JNDI;

    /**
     * Name of the built-in MongoDB probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_MONGO_DB = AgentConstants.PROBE_NAME_MONGO_DB;

    /**
     * Name of the built-in Cassandra probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_CASSANDRA = AgentConstants.PROBE_NAME_CASSANDRA;

    /**
     * Name of the built-in Cassandra probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_HBASE = AgentConstants.PROBE_NAME_HBASE;

    /**
     * Name of the built-in RMI probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_RMI = AgentConstants.PROBE_NAME_RMI;

    /**
     * Name of the built-in web service probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_WS = AgentConstants.PROBE_NAME_WS;

    /**
     * Name of the built-in servlet probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_SERVLET = AgentConstants.PROBE_NAME_SERVLET;

    /**
     * Name of the built-in file probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_FILE = AgentConstants.PROBE_NAME_FILE;

    /**
     * Name of the built-in socket probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_SOCKET = AgentConstants.PROBE_NAME_SOCKET;

    /**
     * Name of the built-in process probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_PROCESS = AgentConstants.PROBE_NAME_PROCESS;

    /**
     * Name of the built-in class loader probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_CLASS_LOADER = AgentConstants.PROBE_NAME_CLASS_LOADER;

    /**
     * Name of the built-in Exception probe. Used in {@link #startProbeRecording(String, boolean)} and {@link #stopProbeRecording(String)}.
     */
    public static final String PROBE_NAME_EXCEPTION = AgentConstants.PROBE_NAME_EXCEPTION;

    /**
     * Starts recording CPU data.
     * This method can be called repeatedly and alternatingly with
     * {@code stopCPURecording()}. With these methods you can restrict
     * CPU profiling to certain regions of your code.
     *
     * @param reset if {@code true}, any previously accumulated CPU profiling
     *              data will be discarded. If {@code false}, CPU data will be accumulated
     *              across pairs of invocations of  {@code startCPURecording()} and
     *              {@code stopCPURecording()}.
     */
    public synchronized static void startCPURecording(boolean reset) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.startCPURecording(reset);
        }
    }

    /**
     * Starts recording CPU data.
     * This method can be called repeatedly and alternatingly with
     * {@code stopCPURecording()}. With these methods you can restrict
     * CPU profiling to certain regions of your code.
     *
     * @param reset if {@code true}, any previously accumulated CPU profiling
     *              data will be discarded. If {@code false}, CPU data will be accumulated
     *              across pairs of invocations of  {@code startCPURecording()} and
     *              {@code stopCPURecording()}.
     * @param trackingOptions an array of TRACKING_ constants or {@code null} for the previous state
     */
    public synchronized static void startCPURecording(boolean reset, TrackingOptions trackingOptions) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.startCPURecording(reset, trackingOptions);
        }
    }

    /**
     * Start recording data for a selected probe. For more information and a more general signature see
     * {@link #startProbeRecording(String, ProbeRecordingOptions)}.
     */
    public static void startProbeRecording(String probeName, boolean events) {
        startProbeRecording(probeName, events ? ProbeRecordingOptions.EVENTS : ProbeRecordingOptions.NONE);
    }

    /**
     * Start recording data for a selected probe.
     * @param probeName the name of the probe. For built-in probes, see the {@code PROBE_NAME} constants in this class.
     * For custom probes, this name is {@code custom.n} where n is the one-based index of the custom probe.
     * @param recordingOptions determines whether special recordings for the probe should be switched on
     * @see #stopProbeRecording(String)
     * @see ProbeRecordingOptions
     */
    public static void startProbeRecording(String probeName, ProbeRecordingOptions recordingOptions) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            boolean[] specialRecording = null;
            if (probeName.equals(AgentConstants.PROBE_NAME_JDBC)) {
                specialRecording = new boolean[] {recordingOptions.isConnectionLeaks()};
            }
            ControllerImpl.startProbeRecording(probeName, recordingOptions.isEvents(), specialRecording);
        }
    }

    /**
     * Stop recording data for a selected probe.
     * @param probeName the name of the probe. For built-in probes, see the {@code PROBE_NAME} constants in this class.
     * For custom probes, this name is {@code custom.n} where n is the one-based index of the custom probe.
     * @see #startProbeRecording(String, boolean)
     */
    public static void stopProbeRecording(String probeName) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.stopProbeRecording(probeName);
        }
    }

    /**
     * Start tracking selected elements for a selected probe. Probe recording must be switched on for the selected
     * probe otherwise this method does not have any effect.
     * @param probeName the name of the probe. For built-in probes, see the {@code PROBE_NAME} constants in this class.
          * For custom probes, this name is {@code custom.n} where n is the one-based index of the custom probe.
     * @param descriptions the descriptions of the tracked elements. These are the strings that you see in the probe view in the JProfiler GUI.
     *                     The sum of the selected elements will be tracked in a single graph. To get separate graphs,
     *                     call this method multiple times.
     * @param objectType the type of the elements to be tracked
     * @param valueType the type of the values to be tracked
     */
    public static void startProbeTracking(String probeName, String[] descriptions, ProbeObjectType objectType, ProbeValueType valueType) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.startProbeTracking(probeName, descriptions, objectType, valueType);
        }
    }

    /**
     * Stop tracking selected elements for a selected probe. This only has an effect if
     * {@link #startProbeTracking(String, String[], ProbeObjectType, ProbeValueType)} has been called before.
     * @param probeName the name of the probe. For built-in probes, see the {@code PROBE_NAME} constants in this class.
               * For custom probes, this name is {@code custom.n} where n is the one-based index of the custom probe.
     * @param description the descriptions of the tracked elements
     * @param objectType the type of the elements to be tracked
     * @param valueType the type of the values to be tracked
     */
    public static void stopProbeTracking(String probeName, String[] description, ProbeObjectType objectType, ProbeValueType valueType) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.stopProbeTracking(probeName, description, objectType, valueType);
        }
    }

    /**
     * Stops CPU recording.
     * This method can be called after {@code startCPURecording()} has been called.
     * However, you do not have to call it since CPU profiling can run until the
     * JVM exits.
     */
    public synchronized static void stopCPURecording() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.stopCPURecording();
        }
    }

    /**
     * Starts the call tracer.
     * This method can be called repeatedly and alternatingly with
     * {@code stopCallTracer()}.
     *
     * @param recordFiltered if {@code true}, calls into filtered classes will be recorded, too.
     * @param cap the maximum number of events to be recorded. A good default is 100000.
     * @param reset if {@code true} previously recorded calls will be cleared..
     */
    public synchronized static void startCallTracer(int cap, boolean recordFiltered, boolean reset) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.startCallTracer(cap, recordFiltered, reset);
        }
    }

    /**
     * Stops the call tracer.
     * This method can be called after {@code startCallTracer()} has been called.
     * However, you do not have to call it since the call tracer will stop automatically
     * after the cap has been reached. The data will be delivered to the frontend afterwards
     * if used in online mode.
     */
    public synchronized static void stopCallTracer() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.stopCallTracer();
        }
    }

    /**
     * Starts method statistics recording.
     * This will reset previously recorded method statistics.
     * This method can be called repeatedly and alternatingly with
     * {@code stopMethodStatsRecording()}.
     */
    public static void startMethodStatsRecording() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.startMethodStatsRecording();
        }
    }

    /**
     * Stops method statistics recording.
     */
    public static void stopMethodStatsRecording() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.stopMethodStatsRecording();
        }
    }

    /**
     * Start recording of monitor usage with default thresholds of 100 microseconds
     * for blocking events and 100 ms for waiting events.
     * This method can be called repeatedly and alternatingly with
     * {@code stopMonitorRecording()}. Monitor profiling is switched off
     * by default.
     */
    public static void startMonitorRecording() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.startMonitorRecording();
        }
    }

    /**
     * Start recording of monitor usage.
     * This method can be called repeatedly and alternatingly with
     * {@code stopMonitorRecording()}. Monitor profiling is switched off
     * by default.
     * @param blockingThreshold the recording threshold for blocking events in microseconds.
     * @param waitingThreshold the recording threshold for waiting events in microseconds.
     */
    public static void startMonitorRecording(int blockingThreshold, int waitingThreshold) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.startMonitorRecording(blockingThreshold, waitingThreshold);
        }
    }

    /**
     * Stop recording of monitor usage.
     * This method can be called repeatedly and alternatingly with
     * {@code startMonitorRecording()}. However, you do not have to call it since
     * monitor profiling can run until the JVM exits.
     */
    public static void stopMonitorRecording() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.stopMonitorRecording();
        }
    }

    /**
     * Starts recording of memory allocations.
     * This method can be called repeatedly and alternatingly with
     * {@code stopAllocRecording()}. With these methods you can restrict
     * memory allocation profiling to certain regions of your code.
     * This is especially useful for profiling an application running within
     * an application server.
     *
     * @param reset if {@code true}, any previously recorded profiling
     *              data will be discarded. If {@code false}, allocations within all
     *              pairs of invocations of  {@code startAllocRecording()} and
     *              {@code stopAllocRecording()} will be recorded.
     */
    public synchronized static void startAllocRecording(boolean reset) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.startAllocRecording(reset);
        }
    }

    /**
     * Starts recording of memory allocations.
     * This method can be called repeatedly and alternatingly with
     * {@code stopAllocRecording()}. With these methods you can restrict
     * memory allocation profiling to certain regions of your code.
     * This is especially useful for profiling an application running within
     * an application server.
     *
     * @param reset if {@code true}, any previously recorded profiling
     *              data will be discarded. If {@code false}, allocations within all
     *              pairs of invocations of  {@code startAllocRecording()} and
     *              {@code stopAllocRecording()} will be recorded.
     * @param trackingOptions the request tracking options
     */
    public synchronized static void startAllocRecording(boolean reset, TrackingOptions trackingOptions) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.startAllocRecording(reset, trackingOptions);
        }
    }


    /**
     * Stops recording of memory allocations.
     * This method can be called after {@code startAllocRecording()} has been called.
     * However, you do not have to call it since memory profiling can run until the
     * JVM exits.
     */
    public synchronized static void stopAllocRecording() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.stopAllocRecording();
        }
    }

    /**
     * Adds a bookmark at the current time.
     * The bookmark will be displayed in all JProfiler graphs with a time axis.
     * The description will be displayed in the tooltip for the bookmark.
     *
     * @param description the name of the bookmark, may also be {@code null}
     */
    public synchronized static void addBookmark(String description) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.addBookmark(description);
        }
    }

    /**
     * Add a bookmark at the current time.
     * The bookmark will be displayed in all JProfiler graphs with a time axis.
     * The description will be displayed in the tooltip for the bookmark.
     * @param description the name of the bookmark, may also be {@code null}
     * @param color the color to be used for drawing the bookmark. If {@code null} the default color will be used.
     * @param dashed if the line for drawing the bookmark should be dashed or not
     */
    public synchronized static void addBookmark(String description, BookmarkColor color, boolean dashed) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.addBookmark(description, color == null ? null : color.getRGB(), dashed);
        }
    }

    /**
     * Trigger a thread dump.
     */
    public synchronized static void triggerThreadDump() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.triggerThreadDump();
        }
    }    

    /**
     * Trigger a heap dump.
     * Calls {@code triggerHeapDump(HeapDumpOptions.DEFAULT)}.
     * @see #triggerHeapDump(HeapDumpOptions)
     */
    public synchronized static void triggerHeapDump() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            triggerHeapDump(HeapDumpOptions.DEFAULT);
        }
    }

    /**
     * Trigger a heap dump with the specified {@link HeapDumpOptions heap dump options}.
     * If you want to analyze a heap dump with the heap walker in a snapshot saved with
     * the {@code saveSnapshot} method, you should call this method from your source code
     * at an appropriate time.
     * <p>
     * <b>ATTENTION:</b> Taking a heap dump takes a long time (on the order of seconds).
     * If you call this method to often, your application might become unusable or take
     * an excessively long time to finish.
     * </p>
     * @param heapDumpOptions the options for the heap dump
     */
    public synchronized static void triggerHeapDump(HeapDumpOptions heapDumpOptions) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.triggerHeapDump(heapDumpOptions);
        }
    }

    /**
     * Saves a snapshot of all profiling data to disk.
     * This is especially important for offline profiling.
     * You should choose the standard extension {@code .jps} for the
     * {@code file} parameter, since JProfiler's GUI frontend filters the corresponding
     * file choosers for that extension. If you want to save several snapshots during
     * one profiling run, please take care to provide unique {@code file} parameters
     * since snapshot files will be overwritten otherwise.
     * <p>
     * <b>ATTENTION:</b> Saving a snapshot takes a long time (on the order of seconds).
     * If you call this method to often, your application might become unusable or take
     * an excessively long time to finish, and your hard disk might run out of space.
     * </p>
     *
     * @param file the file to which the snapshot should be saved.
     */
    public synchronized static void saveSnapshot(final File file) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.saveSnapshot(file, false);
        }
    }

    /**
     * Starts recording of thread states and monitor usage.
     * This method can be called repeatedly and alternatingly with
     * {@code stopThreadProfiling()}. For an offline session, thread
     * profiling is switched on by default.
     */
    public static void startThreadProfiling() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.startThreadProfiling();
        }
    }

    /**
     * Stops recording of thread states and monitor usage.
     * This method can be called repeatedly and alternatingly with
     * {@code startThreadProfiling()}. However, you do not have to call it since
     * thread profiling can run until the JVM exits.
     */
    public static void stopThreadProfiling() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.stopThreadProfiling();
        }
    }

    /**
     * Starts recording of VM telemetry data.
     * This method can be called repeatedly and alternatingly with
     * {@code stopVMTelemetryRecording()}. For an offline session, VM telemetry
     * recording is switched on by default.
     */
    public static void startVMTelemetryRecording() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.startVMTelemetryRecording();
        }
    }

    /**
     * Stops recording of VM telemetry data.
     * This method can be called repeatedly and alternatingly with
     * {@code startVMTelemetryRecording()}. However, you do not have to call it since
     * VM telemetry recording can run until the JVM exits.
     */
    public static void stopVMTelemetryRecording() {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.stopVMTelemetryRecording();
        }
    }

    /**
     * Saves a snapshot of all profiling data to disk when the VM shuts down.
     * This is especially important for offline profiling.
     * You should choose the standard extension {@code .jps} for the
     * {@code file} parameter, since JProfiler's GUI frontend filters the corresponding
     * file choosers for that extension.
     * <p>
     * <b>ATTENTION:</b> Saving a snapshot can take quite some time (on the order of seconds).
     * When the VM is shut down during a user logout or a system shutdown, the OS may terminate
     * the VM before saving is completed.
     *
     * @param file the file to which the snapshot should be saved.
     */
    public static void saveSnapshotOnExit(final File file) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.saveSnapshotOnExit(file);
        }
    }

    /**
     * Enable or disable all triggers with a specified group ID. The group ID can be entered in
     * the "Group ID" step of the trigger configuration wizard in the JProfiler GUI.
     * @param enabled if the triggers should be enabled
     * @param groupId the group ID
     * @throws IllegalArgumentException if no trigger with the specified group ID exists
     */
    public static void enableTriggerGroup(boolean enabled, String groupId) throws IllegalArgumentException {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.enableTriggerGroup(enabled, groupId);
        }
    }

    /**
     * Enable or disable all triggers. The enabled/disabled state of the single triggers will not be lost,
     * disabling all triggers with this method overrides the enabled/disabled state of the single triggers.
     * @param enabled if the triggers should be enabled
     */
    public static void enableTriggers(boolean enabled) {
        if (!ControllerImpl.isPublicControllerDisabled()) {
            ControllerImpl.enableTriggers(enabled);
        }
    }

    private Controller() {
    }

    /**
     * Color class for bookmarks. A custom class is used to avoid the dependency to the AWT classes.
     **/
    public static class BookmarkColor {
        private int value;

        /**
         * Creates an opaque color with the specified red, green,
         * and blue values in the range (0 - 255).
         * This is equivalent to the constructor of {@link java.awt.Color#Color(int, int, int)}
         *
         * @param r the red component
         * @param g the green component
         * @param b the blue component
         */
        @SuppressWarnings("PointlessBitwiseExpression")
        public BookmarkColor(int r, int g, int b) {
            value = ((255 & 0xFF) << 24) |
                    ((r & 0xFF) << 16) |
                    ((g & 0xFF) << 8)  |
                    ((b & 0xFF) << 0);
        }

        private int getRGB() {
            return value;
        }
    }
}
